(* Ocsigen
 * http://www.ocsigen.org
 * Module Eliom_duce
 * Copyright (C) 2007 Vincent Balat, Alain Frisch
 * CNRS - Université Paris Diderot Paris 7
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, with linking exception;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)
open Ocsigen_http_frame
open Ocsigen_http_com
open Lwt
open Ocsigen_senders
open Ocsigen_stream
open Ocsigen_extensions
(*open Eliom_mkforms*)
open Eliom_mkreg

module CT = Cduce_lib.Types

let print_xml xml =
  Cduce_lib.Print_xml.print_xml ~utf8:true !Cduce_lib.Eval.ns_table xml;;

let to_string = Cduce_lib.Value.get_string_latin1
  
(* a reecrire en cduce *)
(*
let add_css (a : html) : html =
  let css =
    {{ <style type="text/css">"\n.eliom_inline {display: inline}\n.eliom_nodisplay {display: none}\n"}}
  in
  {{ match a with <html (al)>el ->
   <html (al)>
      (map* el with <head (al)> el -> [ <head (al)> [ css !el ] ])
   }}
*)

let add_css a = a

let code_of_code_option = function
  | None -> 200
  | Some c -> c

module Cduce_content =
  struct
    type t = Cduce_lib.Value.t

    let get_etag_aux x =
      Some (Digest.to_hex (Digest.string x))

    let print x =
      to_string(print_xml x)

    let get_etag c =
      let x = print (add_css c) in
      get_etag_aux x

(*
    let doctype =
      "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n\n"

    let header_length =
      String.length doctype + String.length Ocsigen_pervasives.advert
*)      
    module S = Ocsigen_stream.StringStream

    let result_of_content c =
      let x = print (add_css c) in
      let md5 = get_etag_aux x in
      let default_result = default_result () in
      Lwt.return
        {default_result with
         res_content_length =
            Some (Int64.of_int (String.length x));
         res_content_type = Some "text/html";
         res_etag = md5;
         res_headers= Http_headers.dyn_headers;
         res_stream =
             (S.make (S.put x), None)
       }
  end

module Xhtmlreg_ = struct
  type page = Cduce_lib.Value.t
  type options = unit

  type return = Eliom_services.http

  let pre_service ?options () = Lwt.return ()

  let send_appl_content = Eliom_services.XNever

  let send ?options ?charset ?code
      ?content_type ?headers content =
    Cduce_content.result_of_content content >>= fun r ->
    Lwt.return
      {r with
         res_cookies= (Eliom_request_info.get_user_cookies ());
         res_code= code_of_code_option code;
         res_charset= Some "utf-8" (* For Eliom_duce, we impose utf-8 *);
         res_content_type= (match content_type with
                              | None -> r.res_content_type
                              | _ -> content_type
                           );
         res_headers= (match headers with
                         | None -> r.res_headers
                         | Some headers -> 
                             Http_headers.with_defaults 
                               headers r.res_headers
                      );
      }
end

module Xhtmlreg = Eliom_mkreg.MakeRegister(Xhtmlreg_)

module Xhtml = struct
  include Xhtmlreg
(*  include Xhtmlforms *)
end

module type XhtmlSig =
sig
  include Eliom_mkreg.ELIOMREGSIG with type options = unit and type
      return = Eliom_services.http
  (*  include XhtmlFormsSig *)
end

module SubXhtml =
  functor(T : sig
    type content
    val print : (string -> unit ) -> content -> unit
  end) ->
    (struct      
      module Cont_content =
      struct
        type t = T.content
	    
        let get_etag_aux x =
          Some (Digest.to_hex (Digest.string x))
	    
        let print x =
          let b = Buffer.create 256 in
          T.print (Buffer.add_string b) x;
          Buffer.contents b

        let get_etag c =
          let x = print c in
          get_etag_aux x

	module S = Ocsigen_stream.StringStream
	let (++) = S.concat

        let result_of_content c =
          let x = print c in
          let md5 = get_etag_aux x in
          let default_result = default_result () in
          Lwt.return
            {default_result with
               res_content_length = Some (Int64.of_int (String.length x));
               res_content_type = Some "text/html";
               res_etag = md5;
               res_headers= Http_headers.dyn_headers;
               res_stream = (S.make (S.put x), None)
           }

      end
	
      module Contreg_ = struct
	type page = T.content
	type options = unit	    
	type return = Eliom_services.http

	let pre_service ?options () = Lwt.return ()
	let send_appl_content = Eliom_services.XNever

	let send ?options ?charset ?code
            ?content_type ?headers content =
          Cont_content.result_of_content content >>= fun r ->
            Lwt.return
              {r with
		res_cookies= (Eliom_request_info.get_user_cookies ());
		res_code= code_of_code_option code;
		res_charset= Some "utf-8" 
		  (* For Eliom_duce, we impose utf-8 *);
		res_content_type= (match content_type with
                  | None -> r.res_content_type
                  | _ -> content_type
                );
		res_headers= (match headers with
                  | None -> r.res_headers
                  | Some headers -> 
                      Http_headers.with_defaults
                        headers r.res_headers
                );
              }
      end
      module Contreg = Eliom_mkreg.MakeRegister(Contreg_)
	
(* include Xhtmlforms *)
      include Contreg
    end : XhtmlSig with type page = T.content
    )

module Xml = SubXhtml(struct
  type content = Cduce_lib.Value.t
  let print f x = f (to_string(print_xml x))
end)

type params = { path : string list; serviceName : string};;
type post_params = { postServiceName : string};;
type coservice_params = { coServiceName : string}

let value_to_string v =
 let b = Buffer.create 1024 in
 let ppf = Format.formatter_of_buffer b in
 Cduce_lib.Value.print ppf v;
 Format.pp_print_flush ppf ();
 Buffer.contents b

let trace msg =
  output_string stderr (msg ^ "\n"); flush stderr;;

let make_int i = trace ("make_int:" ^ (string_of_int i));
  Cduce_lib.Value.ocaml2cduce_int i;;
let make_int64 i = Cduce_lib.Value.ocaml2cduce_int64 i;;
let make_float f = Cduce_lib.Value.float f;;
let make_string s = Cduce_lib.Value.string_latin1 s;;
let make_bool b = Cduce_lib.Value.vbool b;;
let nil = Cduce_lib.Value.nil;;
let make_pair x y = Cduce_lib.Value.Pair(x,y);;
let make_string_pair (s1,s2) = make_pair (make_string s1) 
  (make_pair (make_string s2) nil);; 
let make_cduce_option v =
  trace "make_cduce_option";
  make_pair v nil;;
let make_tag s = Cduce_lib.Value.atom_ascii s;;
let no_attr = Cduce_lib.Value.Record Cduce_lib.Imap.empty;;
let make_xml tag content = trace ("<" ^ tag ^ ">[" ^ (value_to_string content) ^ "]"); 
  Cduce_lib.Value.xml (make_tag tag) no_attr content;;
let make_xml_attr tag attr content = 
  trace ("<" ^ tag ^ " " ^ (value_to_string attr) ^ ">[" ^ (value_to_string content) ^ "]"); 
  Cduce_lib.Value.xml (make_tag tag) attr content;;
let rec make_string_list = function
  | [] -> nil
  | h::r -> make_pair (make_string h) (make_string_list r);;
let rec make_list (f : ('a -> Cduce_lib.Value.t)) = function
  | [] -> nil
  | h::r -> make_pair (f h) (make_list f r);;
let rec make_string_pair_list = function
  | [] -> nil
  | h::r -> make_pair (make_string_pair h) (make_string_pair_list r);;

let rec convert_params_type (get_params : CT.service_params) 
    = trace ("convert_params_type:" ^ (CT.Service.to_string get_params));
  match get_params with
	CT.TUnit -> Obj.magic (Eliom_parameters.unit)
      | CT.TProd (t1, t2) -> Obj.magic (Eliom_parameters.prod 
	  (convert_params_type t1)
	  (convert_params_type t2)) 
      | CT.TSum (t1, t2) -> Obj.magic (Eliom_parameters.sum 
	  (convert_params_type t1)
	  (convert_params_type t2)) 
      | CT.TInt n -> Obj.magic (Eliom_parameters.int n)
      | CT.TFloat n -> Obj.magic (Eliom_parameters.float n)
      | CT.TString n -> Obj.magic (Eliom_parameters.string n)
      | CT.TBool n -> Obj.magic (Eliom_parameters.bool n)
      | CT.TOption n -> Obj.magic (Eliom_parameters.opt
	  (convert_params_type n))
      | CT.TSuffix (b,t) -> Obj.magic (Eliom_parameters.suffix 
	  ~redirect_if_not_suffix:b
	  (convert_params_type t))
      | CT.TConst n -> Obj.magic (Eliom_parameters.suffix_const n)
      | CT.TESuffix n -> Obj.magic (Eliom_parameters.all_suffix n)
      | CT.TList (n,t) -> Obj.magic (Eliom_parameters.list n
	  (convert_params_type t))
      | CT.TAny -> Obj.magic (Eliom_parameters.any)
      | CT.TFile n -> trace ("debug:TFile:" ^ n); Obj.magic (Eliom_parameters.file n)
(*      | CT.TSet n -> Obj.magic (Eliom_parameters.set
	(convert_params_type n)) ??? *)
      | _ -> assert false;;

let rec process_params (param_type : CT.service_params) (param : 'a)
    = trace ("debug process_params:"^(CT.Service.to_string param_type));
  match param_type with 
    | CT.TProd (CT.TProd (t'1,t'2), t2) 
      -> trace ("debug process_params cas 1:t'1="
      ^(CT.Service.to_string t'1 ));
	(match (Obj.magic param) with 
	    ((p'1,p'2),p2) -> 	trace "debug process_params cas 1.1";
	      let (ptype, pval) = process_params (CT.TProd(t'2,t2))
		(p'2,p2) in 
	      (CT.TProd (t'1,ptype), Obj.magic (p'1, pval)))
    | _ -> trace "debug process_params cas 2";
	(param_type, Obj.magic param);;
	      
let convert_base_type (param_type : CT.service_params) (param : 'a) =
  trace ("convert_base_type:" ^ (CT.Service.to_string param_type));
  match param_type with 
    | CT.TInt _ -> make_int (Obj.magic param : int)
    | CT.TFloat _ -> make_float (Obj.magic param : float)
    | CT.TString _ -> make_string (Obj.magic param : string)
    | CT.TBool _ -> make_bool (Obj.magic param : bool)
    | _ -> assert false;;

let make_base_xml (n : string) t =
  Obj.magic (make_xml n (make_pair t nil));;

let rec convert_params_ (param_type : CT.service_params) (param : 'a) 
    = trace ("debug convert_params_"^(CT.Service.to_string param_type));
  match param_type with
	CT.TUnit -> nil
      | CT.TProd (t1, t2) -> 
	  (match (Obj.magic param) with 
	      (p1,p2) -> (match t2 with
		| CT.TProd _ ->  Obj.magic (make_pair 
		    (convert_params_ t1 p1) 
		    (convert_params_ t2 p2))
		| _ -> Obj.magic (make_pair 
		    (convert_params_ t1 p1) 
		    (make_pair (convert_params_ t2 p2) nil)))
	  )
      | CT.TSum (t1, t2) -> 
	  (match (Obj.magic param) with 
	    | Eliom_parameters.Inj1 p -> convert_params_ t1 p
	    | Eliom_parameters.Inj2 p -> convert_params_ t2 p)
      | CT.TOption t -> trace ("convert_params option:" ^
	  (CT.Service.to_string t));
	  let value = (match (Obj.magic param) with 
	    | Some v -> make_pair (convert_base_type t v) nil
	    | None -> nil) in
	  let name = (match t with 
	    | CT.TInt n | CT.TFloat n | CT.TString n | CT.TBool n
		-> n
	    | _ -> assert false) in
	  Obj.magic (make_xml name value)
      | CT.TInt n | CT.TFloat n | CT.TString n | CT.TBool n
	  -> make_base_xml n (convert_base_type param_type param)
      | CT.TFile n ->
	let info = (Obj.magic param : Ocsigen_extensions.file_info) in
	let tmp_name = Eliom_request_info.get_tmp_filename info in
	let orig_name = Eliom_request_info.get_original_filename info in
	let size = Eliom_request_info.get_filesize info in
	let content = make_string tmp_name in
	let attr_list = [("orig_name",(make_string orig_name));
			 ("size", (make_int64 size)) ] in
	let attrs = Cduce_lib.Value.record_ascii attr_list in
	make_xml_attr n attrs (make_pair content nil)
      | CT.TConst n -> Obj.magic (make_xml n (make_pair (make_string n) nil))
      | CT.TESuffix n -> Obj.magic (make_xml n (make_pair
	  (make_string_list (Obj.magic param : string list)) nil))
      | CT.TSuffix (_, t) ->
	(let content = convert_params_ t param in
	match content with
	  | Cduce_lib.Value.Pair _ -> 
	    make_xml "eliom:suffix" content
	  | _ -> 
	    make_base_xml "eliom:suffix" content)
      | CT.TList (n,t) ->
	  Obj.magic (make_xml n (make_list (convert_params_ t)
	    (Obj.magic param : 'a list)))
      | CT.TAny -> Obj.magic (make_xml "eliom:any" 
				(make_pair
				   (make_string_pair_list
				      (Obj.magic param : (string * string) list)) nil))
      | _ -> assert false;;

let convert_params param_type param =
  let is_compound = function 
    | Cduce_lib.Value.Pair _ -> true
    | _ -> false
  in
  let (ptype,pvalue) = process_params param_type param in
  let value = convert_params_ ptype pvalue  in
  trace ("debug:convert_params:" ^ (value_to_string value));
  if (is_compound value) then 
    (trace ("debug:convert_params:compound:" ^ (value_to_string value)); value)
  else
    (trace ("debug:convert_params:!compound:" ^ (value_to_string (make_pair
  value nil)));
    make_pair value nil);;

let service_table = Hashtbl.create 10;;

let get_service (id : string) =
  trace ("get_service " ^ id);
  Hashtbl.find service_table id;;

let put_service serviceID service =
  Hashtbl.add service_table serviceID service;
  trace ("put_service " ^ serviceID);;

let get_fun abstraction =
  match abstraction with
    | Cduce_lib.Value.Abstraction (_,f) -> 
	f
    | _ -> assert false;;
 
let send_id id res =
  trace ("send_id " ^ (string_of_int id));
  match res with
    | Cduce_lib.Value.Abstraction (_,f') -> 
	(* more parameters are expected -> send ID *)
	f' (make_int id)
    | _ -> res;;

let convert_function_params (function_params : CT.service_params) =
  let param = convert_params_type function_params in
  trace ("param reg:" ^ (CT.Service.to_string function_params));
  (function_params,param);;

let fun_get_parameters page_gen rep_p get_p () =
  (*  let id = put_server_params (sp,get_p) in *)
  let tmp = page_gen (Obj.magic(convert_params rep_p get_p)) in
  (* let res = send_id (* id *) tmp in *)
  (*  remove_server_params id; *)
  Lwt.return tmp;;

let register_cduce_service_ (service_params : params)
    (function_params : CT.service_params)
    (page_gen : ('a -> Cduce_lib.Value.t)) =
  trace ("register_cduce_service_:"^ service_params.serviceName ^ "="
    ^ (List.hd service_params.path));
  let (rfp,fp) = convert_function_params function_params in
  let serviceID = service_params.serviceName in
  let service = Xhtml.register_service ~path:service_params.path
    ~get_params:fp
    (fun_get_parameters page_gen rfp) in
  put_service serviceID (service,fp,rfp);;

let print_arg (i,j) =
  output_string stderr ("argument = (" ^
    (Cduce_lib.Types.Print.to_string i) ^  " -> " ^  
    (Cduce_lib.Types.Print.to_string
      j) ^ ")\n");
  output_string stderr ("-----------------\n");;
 
let get_abstraction_args a =
  match a with 
    | Cduce_lib.Value.Abstraction (args,f) ->
	(match args with
	  | None -> assert false
	  | Some la -> (la,f))
    | _ -> assert false;;

let get_pagegen_arg (args,args') =
  print_arg (args,args');
  CT.Service.to_service_params args;;

let get_pagegen_args a = 
  match a with 
    | None -> [ CT.TUnit ]
    | Some la -> List.map get_pagegen_arg la;;

let convert_fun_param p =
  match p with 
    | Cduce_lib.Value.Abstraction (args,_ (* function *) ) ->
	(match args with
	  | None -> trace "no parameter"
	  | Some la -> print_arg (List.hd la))
    | _ -> assert false;;

let get_pagegen_function p =
  match p with
    | Cduce_lib.Value.Abstraction (args, f) ->
	((get_pagegen_args args),f)
    | _ -> assert false;;
 
let register_arg sp f args =
  register_cduce_service_ sp args f;;

let register_cduce_service (service_params : params) 
    (page_gen : Cduce_lib.Value.t) =
  output_string stderr ("Path=" ^ (List.hd service_params.path) ^ "\n");
  let (args,f) = get_pagegen_function page_gen in
  List.iter (register_arg service_params f) args;;

let register_cduce_post_service_ (service_params : post_params)
    (page_gen : ('a -> Cduce_lib.Value.t)) 
    (fallback_id : string) 
    (function_post_params : CT.service_params) =
  trace ("register_cduce_post_service_:" ^ service_params.postServiceName);
  let (rpostfp,postfp) = convert_function_params function_post_params in
  let serviceID = service_params.postServiceName in
  let (fallback_service, _, function_get_params) = 
    get_service fallback_id in
  trace ("get reg:" ^ (CT.Service.to_string function_get_params));
  let service = Xhtml.register_post_service 
    ~fallback:fallback_service
    ~post_params:postfp
    (fun get_p post_p ->
      let tmp_fun = page_gen (Obj.magic(convert_params rpostfp
					  post_p)) in
      let tmp = (get_fun tmp_fun) (Obj.magic(convert_params function_get_params get_p)) in
      return tmp) in
  put_service serviceID (Obj.magic(service),postfp,rpostfp);;

let register_post_arg sp f f_id args =
  register_cduce_post_service_ sp f f_id args;;

let register_cduce_post_service (service_params : post_params) 
    (fallback_id : string) (page_gen : Cduce_lib.Value.t) =
  output_string stderr ("Name=" ^ (service_params.postServiceName) ^ "\n");
  let (args,f) = get_pagegen_function page_gen in
  List.iter (register_post_arg service_params f fallback_id) args;;
  
let register_cduce_coservice_ (params : coservice_params)
    (page_gen : ('a -> Cduce_lib.Value.t)) 
    (fallback_id : string) 
    (function_get_params : CT.service_params) =
  trace ("register_cduce_coservice_:" ^ params.coServiceName);
  let (rgetfp,getfp) = convert_function_params function_get_params in
  let serviceID = params.coServiceName in
  let (fallback_service, _, _) = 
    get_service fallback_id in (* A verifier: seulement unit service
				  sont autorises *)
  let service = Xhtml.register_coservice
    ~fallback:(Obj.magic (fallback_service))
    ~get_params:getfp
    (fun_get_parameters page_gen rgetfp) in
  put_service serviceID (service,getfp,rgetfp);;

let register_coservice_arg sp f f_id args =
  register_cduce_coservice_ sp f f_id args;;

let register_cduce_coservice (service_params : coservice_params) 
    (fallback_id : string) (page_gen : Cduce_lib.Value.t) =
  let (args,f) = get_pagegen_function page_gen in
  List.iter (register_coservice_arg service_params f fallback_id)
    args;;

let register_cduce_post_coservice_ (params : coservice_params)
    (page_gen : ('a -> Cduce_lib.Value.t)) 
    (fallback_id : string) 
    (function_get_params : CT.service_params) =
  trace ("register_cduce_post_coservice_:" ^ params.coServiceName);
  let (rgetfp,getfp) = convert_function_params function_get_params in
  let serviceID = params.coServiceName in
  let (fallback_service, _, _) = 
    get_service fallback_id in (* A verifier: seulement unit service
				  sont autorises *)
  let service = Xhtml.register_post_coservice
    ~fallback:(Obj.magic (fallback_service))
    ~post_params:getfp
    (fun_get_parameters page_gen rgetfp) in
  put_service serviceID (Obj.magic(service),Obj.magic(getfp),rgetfp);;

let register_post_coservice_arg sp f f_id args =
  register_cduce_post_coservice_ sp f f_id args;;

let register_cduce_post_coservice (service_params : coservice_params) 
    (fallback_id : string) (page_gen : Cduce_lib.Value.t) =
  let (args,f) = get_pagegen_function page_gen in
  List.iter (register_post_coservice_arg service_params f fallback_id)
    args;;

let debug_string msg =
  trace ("CDuce debug:" ^ msg);;

let tracePair (s1,s2) =
  trace ("("^s1^","^s2^")");;

let make_string_base_uri_ (serviceID : string) 
    get_id_fun : string * (string * string) list =
  let (service,_,_) = get_id_fun serviceID in
  trace ("make_string_uri:" ^ serviceID);
  let (uri, pregetparams, fragment) = Eliom_lib.make_uri_components_
    ~service:(Obj.magic service) ()
  in
  trace ("make_string_uri2:" ^ uri);
  List.iter tracePair pregetparams;
  (match fragment with 
    | None -> trace "No fragment"
    | Some s -> trace ("Fragment:"^s));
  (uri,pregetparams);;

let make_string_base_uri (serviceID : string) 
    : string * (string * string) list =
  trace ("make_string_uri:" ^ serviceID);
  let (uri,pg) = make_string_base_uri_ serviceID get_service in
  trace ("make_string_uri2:" ^ uri);
  (uri,pg);;

let make_string_base_post_uri_ (serviceID : string) 
    get_id_fun : string * (string * string) list * (string * string) list =
  let (service,_,_) = get_id_fun serviceID in
  trace ("make_string_post_uri:" ^ serviceID);
  let (uri, params, fragment, postparams) = Eliom_lib.make_post_uri_components_
    ~service:(Obj.magic service) () ()
  in
  trace ("make_string_post_uri2:" ^ uri);
  List.iter tracePair postparams;
  (match fragment with 
    | None -> trace "No fragment"
    | Some s -> trace ("Fragment:"^s));
  (uri,params,postparams);;

let make_string_base_post_uri (serviceID : string) 
    : string * (string * string) list * (string * string) list =
  trace ("make_string_uri:" ^ serviceID);
  let (uri,params,postparams) = make_string_base_post_uri_ serviceID get_service in
  trace ("make_string_uri2:" ^ uri);
  (uri,params,postparams);;
